
// Copyright (C) 2008 Indrek Mandre <indrek@mare.ee>
// http://www.mare.ee/indrek/timer/

#ifndef __xavr_h__
#define __xavr_h__

// this is the start of my code library for programming the AVR family
// XXX: I misunderstood interrupts at my dry no-hw programming
// so all of that is wrong here, the pins, inputs, outputs,
// and leds all work fine though

#if defined(__AVR_ATmega48) || defined(__AVR_ATmega88) || defined(__AVR_ATmega168__)
  #include "iomx8.h"
#else
  #warn "XAVR: Unkown/unsupported MCU"
#endif

#define _XAVR_INLINE __attribute__((always_inline))

#define _XAVR_STR(S) _XAVR_STR2(S)
#define _XAVR_STR2(S) #S
#define _XAVR_CONCAT2(A,B) _XAVR_CONCAT2b(A,B)
#define _XAVR_CONCAT2b(A,B) A ## B
#define _XAVR_CONCAT3(A,B,C) _XAVR_CONCAT3b(A,B,C)
#define _XAVR_CONCAT3b(A,B,C) A ## B ## C

#ifndef _XAVR_PORTA_MASK
  #define _XAVR_PORTA_MASK 0xff
#endif
#ifndef _XAVR_PORTB_MASK
  #define _XAVR_PORTB_MASK 0xff
#endif
#ifndef _XAVR_PORTC_MASK
  #define _XAVR_PORTC_MASK 0xff
#endif
#ifndef _XAVR_PORTD_MASK
  #define _XAVR_PORTD_MASK 0xff
#endif
#ifndef _XAVR_PORTE_MASK
  #define _XAVR_PORTE_MASK 0xff
#endif
#ifndef _XAVR_PORTF_MASK
  #define _XAVR_PORTF_MASK 0xff
#endif
#ifndef _XAVR_PORTG_MASK
  #define _XAVR_PORTG_MASK 0xff
#endif
#ifndef _XAVR_PORTH_MASK
  #define _XAVR_PORTG_MASK 0xff
#endif

namespace XAVR
{

#define _XAVR_PORT_INTERNAL(N) \
  struct _XAVR_Port ## N \
  { \
    enum { PORT_MASK = _XAVR_PORT ## N ## _MASK }; \
    static inline void Init(uint8_t im, uint8_t om) _XAVR_INLINE \
    { \
      if ( im != 0 || om != 0 ) \
        { \
          if ( im == PORT_MASK ) \
            { \
              DDR ## N = 0; \
              PORT ## N = PORT_MASK; \
            } \
          else if ( om == PORT_MASK ) \
            { \
              DDR ## N = PORT_MASK; \
            } \
          else \
            { \
              int8_t v = DDR ## N; \
              v |= om; \
              v &= (uint8_t)~im; \
              DDR ## N = v; \
            } \
        } \
    } \
    static inline void AddM(uint8_t f) _XAVR_INLINE \
    { \
      if ( f == PORT_MASK ) \
          PORT ## N = (uint8_t)PORT_MASK; \
      else if ( f ) \
          PORT ## N |= (uint8_t)f; \
    } \
    static inline void DelM(uint8_t f) _XAVR_INLINE \
    { \
      if ( f == PORT_MASK ) \
          PORT ## N = (uint8_t)0; \
      else if ( f ) \
          PORT ## N &= (uint8_t)~f; \
    } \
    static inline void XorM(uint8_t f) _XAVR_INLINE \
    { \
      if ( f == PORT_MASK ) \
          PORT ## N = (uint8_t)~PORT ## N; \
      else if ( f ) \
          PORT ## N ^= (uint8_t)f; \
    } \
    static inline void Select(uint8_t set, uint8_t clr) _XAVR_INLINE \
    { \
      if ( set == PORT_MASK ) \
          PORT ## N = (uint8_t)PORT_MASK; \
      else if ( clr == PORT_MASK ) \
          PORT ## N = (uint8_t)0; \
      else if ( (set | clr) == PORT_MASK ) \
          PORT ## N = (uint8_t)set; \
      else if ( set != 0 || clr != 0 ) \
        { \
          uint8_t v = PORT ## N; \
          v |= set; \
          v &= ~clr; \
          PORT ## N = v; \
        } \
    } \
    static inline void Set(uint8_t f) _XAVR_INLINE \
    { \
      PORT ## N = f; \
    } \
    static inline uint8_t Get() _XAVR_INLINE \
    { \
      return PORT ## N; \
    } \
  };

#define _XAVR_PORT_DUMMY(N) \
  struct _XAVR_Port ## N \
  { \
    static inline void Init(uint8_t im, uint8_t om) _XAVR_INLINE { } \
    static inline void AddM(uint8_t f) _XAVR_INLINE { } \
    static inline void DelM(uint8_t f) _XAVR_INLINE { } \
    static inline void XorM(uint8_t f) _XAVR_INLINE { } \
    static inline void Select(uint8_t set, uint8_t clr) _XAVR_INLINE { } \
    static inline void Set(uint8_t f) _XAVR_INLINE { } \
    static inline uint8_t Get() _XAVR_INLINE { return 0; } \
  };

#ifdef PORTA
  _XAVR_PORT_INTERNAL(A);
#else
  _XAVR_PORT_DUMMY(A);
#endif

#ifdef PORTB
  _XAVR_PORT_INTERNAL(B);
#else
  _XAVR_PORT_DUMMY(B);
#endif

#ifdef PORTC
  _XAVR_PORT_INTERNAL(C);
#else
  _XAVR_PORT_DUMMY(C);
#endif

#ifdef PORTD
  _XAVR_PORT_INTERNAL(D);
#else
  _XAVR_PORT_DUMMY(D);
#endif

#ifdef PORTE
  _XAVR_PORT_INTERNAL(E);
#else
  _XAVR_PORT_DUMMY(E);
#endif

#ifdef PORTF
  _XAVR_PORT_INTERNAL(F);
#else
  _XAVR_PORT_DUMMY(F);
#endif

#ifdef PORTG
  _XAVR_PORT_INTERNAL(G);
#else
  _XAVR_PORT_DUMMY(G);
#endif

#ifdef PORTG
  _XAVR_PORT_INTERNAL(H);
#else
  _XAVR_PORT_DUMMY(H);
#endif

// I got these interrupts wrong, ignore in this code all interrupt related stuff
#define _XAVR_PCINT_VECTOR_NAME(PORTNAME, PINID) \
    _XAVR_STR(_XAVR_CONCAT3(PCINT, PORT ## PORTNAME ## PINID ## _PCINT, _vect))

#define _XAVR_PORTPIN(PORTNAME, PINID) \
    typedef OutputPin<PORT ## PORTNAME ## PINID> Out ## PINID; \
    struct In ## PINID : public InputPin<PORT ## PORTNAME ## PINID> \
    { \
      enum { PCINT = PORT ## PORTNAME ## PINID ## _PCINT }; \
      static inline void _XAVR_InjectPCInterruptCode () \
      { \
        __asm__ __volatile__( \
            "\n.global " _XAVR_PCINT_VECTOR_NAME(PORTNAME, PINID) "\n" \
            "	.type   " _XAVR_PCINT_VECTOR_NAME(PORTNAME, PINID) ", @function\n" \
            _XAVR_PCINT_VECTOR_NAME(PORTNAME, PINID) ":\n" \
        ); \
      } \
    };

#define XAVR_PORTA 0
#define XAVR_PORTB 1
#define XAVR_PORTC 2
#define XAVR_PORTD 3
#define XAVR_PORTE 4
#define XAVR_PORTF 5
#define XAVR_PORTG 6
#define XAVR_PORTH 7

#define XAVR_PORTA_NAME "A"
#define XAVR_PORTB_NAME "B"
#define XAVR_PORTC_NAME "C"
#define XAVR_PORTD_NAME "D"
#define XAVR_PORTE_NAME "E"
#define XAVR_PORTF_NAME "F"
#define XAVR_PORTG_NAME "G"
#define XAVR_PORTH_NAME "H"

  template<uint8_t BIT, uint8_t PORT>
  struct GenericOutputPin
  {
    enum {
      INP_PORTA_BIT = 0,
      INP_PORTB_BIT = 0,
      INP_PORTC_BIT = 0,
      INP_PORTD_BIT = 0,
      INP_PORTE_BIT = 0,
      INP_PORTF_BIT = 0,
      INP_PORTG_BIT = 0,
      INP_PORTH_BIT = 0,
      OUT_PORTA_BIT = PORT == XAVR_PORTA ? BIT : 0,
      OUT_PORTB_BIT = PORT == XAVR_PORTB ? BIT : 0,
      OUT_PORTC_BIT = PORT == XAVR_PORTC ? BIT : 0,
      OUT_PORTD_BIT = PORT == XAVR_PORTD ? BIT : 0,
      OUT_PORTE_BIT = PORT == XAVR_PORTE ? BIT : 0,
      OUT_PORTF_BIT = PORT == XAVR_PORTF ? BIT : 0,
      OUT_PORTG_BIT = PORT == XAVR_PORTG ? BIT : 0,
      OUT_PORTH_BIT = PORT == XAVR_PORTH ? BIT : 0
    };

    static inline void CheckIsOutputPin() _XAVR_INLINE { }

  private:
    GenericOutputPin();
  };

  template<uint8_t BIT, uint8_t PORT>
  struct GenericInputPin
  {
    enum {
      OUT_PORTA_BIT = 0,
      OUT_PORTB_BIT = 0,
      OUT_PORTC_BIT = 0,
      OUT_PORTD_BIT = 0,
      OUT_PORTE_BIT = 0,
      OUT_PORTF_BIT = 0,
      OUT_PORTG_BIT = 0,
      OUT_PORTH_BIT = 0,
      INP_PORTA_BIT = PORT == XAVR_PORTA ? BIT : 0,
      INP_PORTB_BIT = PORT == XAVR_PORTB ? BIT : 0,
      INP_PORTC_BIT = PORT == XAVR_PORTC ? BIT : 0,
      INP_PORTD_BIT = PORT == XAVR_PORTD ? BIT : 0,
      INP_PORTE_BIT = PORT == XAVR_PORTE ? BIT : 0,
      INP_PORTF_BIT = PORT == XAVR_PORTF ? BIT : 0,
      INP_PORTG_BIT = PORT == XAVR_PORTG ? BIT : 0,
      INP_PORTH_BIT = PORT == XAVR_PORTH ? BIT : 0
    };

    static inline void CheckIsInputPin() _XAVR_INLINE { }

  private:
    GenericInputPin();
  };

  struct NullPin
  {
    static inline void CheckIsInputPin() _XAVR_INLINE { }
    static inline void CheckIsOutputPin() _XAVR_INLINE { }

    enum {
      INP_PORTA_BIT = 0,
      INP_PORTB_BIT = 0,
      INP_PORTC_BIT = 0,
      INP_PORTD_BIT = 0,
      INP_PORTE_BIT = 0,
      INP_PORTF_BIT = 0,
      INP_PORTG_BIT = 0,
      INP_PORTH_BIT = 0,
      OUT_PORTA_BIT = 0,
      OUT_PORTB_BIT = 0,
      OUT_PORTC_BIT = 0,
      OUT_PORTD_BIT = 0,
      OUT_PORTE_BIT = 0,
      OUT_PORTF_BIT = 0,
      OUT_PORTG_BIT = 0,
      OUT_PORTH_BIT = 0
    };

  private:
    NullPin();
  };


  #define _XAVR_OUTPUT_PIN_TYPE(N,PN) \
    template <uint8_t TPID> \
    class OutputPin : public GenericOutputPin<_BV(TPID),PN> \
    { \
    public: \
      static inline void Init() _XAVR_INLINE { DDR ## N |= _BV(TPID); } \
      static inline void Clear() _XAVR_INLINE { PORT ## N &= (uint8_t)~_BV(TPID); } \
      static inline void Set() _XAVR_INLINE { PORT ## N |= _BV(TPID); } \
      static inline bool IsSet() _XAVR_INLINE { return PORT ## N & _BV(TPID); } \
      static inline void Toggle() _XAVR_INLINE { PORT ## N ^= _BV(TPID); } \
    private: \
      OutputPin(); \
    };

  #define _XAVR_INPUT_PIN_TYPE(N,PN) \
    template <uint8_t TPID> \
    class InputPin : public GenericInputPin<_BV(TPID),PN> \
    { \
    public: \
      static inline void Init () _XAVR_INLINE { DDR ## N &= (uint8_t)~_BV(TPID); PORT ## N |= _BV(TPID); } \
      static inline bool IsOn() _XAVR_INLINE { return PIN ## N & _BV(TPID); } \
    private: \
      InputPin(); \
    };

#ifdef PORTA
  class PortA
  {
  public:
    _XAVR_OUTPUT_PIN_TYPE(A,XAVR_PORTA);
    _XAVR_INPUT_PIN_TYPE(A,XAVR_PORTA);
    #ifdef PORTA0
    _XAVR_PORTPIN(A, 0)
    #endif
    #ifdef PORTA1
    _XAVR_PORTPIN(A, 1)
    #endif
    #ifdef PORTA2
    _XAVR_PORTPIN(A, 2)
    #endif
    #ifdef PORTA3
    _XAVR_PORTPIN(A, 3)
    #endif
    #ifdef PORTA4
    _XAVR_PORTPIN(A, 4)
    #endif
    #ifdef PORTA5
    _XAVR_PORTPIN(A, 5)
    #endif
    #ifdef PORTA6
    _XAVR_PORTPIN(A, 6)
    #endif
    #ifdef PORTA7
    _XAVR_PORTPIN(A, 7)
    #endif

  private:
    PortA();
  };
#endif

#ifdef PORTB
  class PortB
  {
  public:
    _XAVR_OUTPUT_PIN_TYPE(B,XAVR_PORTB);
    _XAVR_INPUT_PIN_TYPE(B,XAVR_PORTB);
    #ifdef PORTB0
    _XAVR_PORTPIN(B, 0)
    #endif
    #ifdef PORTB1
    _XAVR_PORTPIN(B, 1)
    #endif
    #ifdef PORTB2
    _XAVR_PORTPIN(B, 2)
    #endif
    #ifdef PORTB3
    _XAVR_PORTPIN(B, 3)
    #endif
    #ifdef PORTB4
    _XAVR_PORTPIN(B, 4)
    #endif
    #ifdef PORTB5
    _XAVR_PORTPIN(B, 5)
    #endif
    #ifdef PORTB6
    _XAVR_PORTPIN(B, 6)
    #endif
    #ifdef PORTB7
    _XAVR_PORTPIN(B, 7)
    #endif

  private:
    PortB();
  };
#endif

#ifdef PORTC
  class PortC
  {
  public:
    _XAVR_OUTPUT_PIN_TYPE(C,XAVR_PORTC);
    _XAVR_INPUT_PIN_TYPE(C,XAVR_PORTC);
    #ifdef PORTC0
    _XAVR_PORTPIN(C, 0)
    #endif
    #ifdef PORTC1
    _XAVR_PORTPIN(C, 1)
    #endif
    #ifdef PORTC2
    _XAVR_PORTPIN(C, 2)
    #endif
    #ifdef PORTC3
    _XAVR_PORTPIN(C, 3)
    #endif
    #ifdef PORTC4
    _XAVR_PORTPIN(C, 4)
    #endif
    #ifdef PORTC5
    _XAVR_PORTPIN(C, 5)
    #endif
    #ifdef PORTC6
    _XAVR_PORTPIN(C, 6)
    #endif
    #ifdef PORTC7
    _XAVR_PORTPIN(C, 7)
    #endif

  private:
    PortC();
  };
#endif

#ifdef PORTD
  class PortD
  {
  public:
    _XAVR_OUTPUT_PIN_TYPE(D,XAVR_PORTD);
    _XAVR_INPUT_PIN_TYPE(D,XAVR_PORTD);
    #ifdef PORTD0
    _XAVR_PORTPIN(D, 0)
    #endif
    #ifdef PORTD1
    _XAVR_PORTPIN(D, 1)
    #endif
    #ifdef PORTD2
    _XAVR_PORTPIN(D, 2)
    #endif
    #ifdef PORTD3
    _XAVR_PORTPIN(D, 3)
    #endif
    #ifdef PORTD4
    _XAVR_PORTPIN(D, 4)
    #endif
    #ifdef PORTD5
    _XAVR_PORTPIN(D, 5)
    #endif
    #ifdef PORTD6
    _XAVR_PORTPIN(D, 6)
    #endif
    #ifdef PORTD7
    _XAVR_PORTPIN(D, 7)
    #endif

  private:
    PortD();
  };
#endif

  template <bool>
  struct _xavr_user_error
  {
    static inline void you_have_same_pin_for_both_input_and_output_at_Pins__Init () _XAVR_INLINE { }
    static inline void PC_interrupt_not_allowed_for_this_pin () _XAVR_INLINE { }
  };

  // in case of erronous input the compiler will use this specialization,
  // resulting in a compile time error due to missing function
  template <>
  struct _xavr_user_error<true>
  {
  };

#define _XAVR_INP_PORT_BITS(N) \
      INP_PORT ## N ## _BITS = \
          A0::INP_PORT ## N ## _BIT | A1::INP_PORT ## N ## _BIT | \
          A2::INP_PORT ## N ## _BIT | A3::INP_PORT ## N ## _BIT | \
          A4::INP_PORT ## N ## _BIT | A5::INP_PORT ## N ## _BIT | \
          A6::INP_PORT ## N ## _BIT | A7::INP_PORT ## N ## _BIT | \
          A8::INP_PORT ## N ## _BIT

#define _XAVR_OUT_PORT_BITS(N) \
      OUT_PORT ## N ## _BITS = \
          A0::OUT_PORT ## N ## _BIT | A1::OUT_PORT ## N ## _BIT | \
          A2::OUT_PORT ## N ## _BIT | A3::OUT_PORT ## N ## _BIT | \
          A4::OUT_PORT ## N ## _BIT | A5::OUT_PORT ## N ## _BIT | \
          A6::OUT_PORT ## N ## _BIT | A7::OUT_PORT ## N ## _BIT | \
          A8::OUT_PORT ## N ## _BIT

  template<class A0, class A1 = NullPin, class A2 = NullPin,
      class A3 = NullPin, class A4 = NullPin, class A5 = NullPin,
      class A6 = NullPin, class A7 = NullPin, class A8 = NullPin>
  class Pins
  {
  public:
    enum
    {
      _XAVR_INP_PORT_BITS(A),
      _XAVR_INP_PORT_BITS(B),
      _XAVR_INP_PORT_BITS(C),
      _XAVR_INP_PORT_BITS(D),
      _XAVR_INP_PORT_BITS(E),
      _XAVR_INP_PORT_BITS(F),
      _XAVR_INP_PORT_BITS(G),
      _XAVR_INP_PORT_BITS(H),

      _XAVR_OUT_PORT_BITS(A),
      _XAVR_OUT_PORT_BITS(B),
      _XAVR_OUT_PORT_BITS(C),
      _XAVR_OUT_PORT_BITS(D),
      _XAVR_OUT_PORT_BITS(E),
      _XAVR_OUT_PORT_BITS(F),
      _XAVR_OUT_PORT_BITS(G),
      _XAVR_OUT_PORT_BITS(H)
    };

    static inline void Init() _XAVR_INLINE
    {
      // generate compile time error in case of erronous input
      _xavr_user_error<(OUT_PORTA_BITS & INP_PORTA_BITS ||
          OUT_PORTB_BITS & INP_PORTB_BITS ||
          OUT_PORTC_BITS & INP_PORTC_BITS ||
          OUT_PORTD_BITS & INP_PORTD_BITS ||
          OUT_PORTE_BITS & INP_PORTE_BITS ||
          OUT_PORTF_BITS & INP_PORTF_BITS ||
          OUT_PORTG_BITS & INP_PORTG_BITS ||
          OUT_PORTH_BITS & INP_PORTH_BITS) ? true : false>::you_have_same_pin_for_both_input_and_output_at_Pins__Init();
      
      _XAVR_PortA::Init(INP_PORTA_BITS, OUT_PORTA_BITS);
      _XAVR_PortB::Init(INP_PORTB_BITS, OUT_PORTB_BITS);
      _XAVR_PortC::Init(INP_PORTC_BITS, OUT_PORTC_BITS);
      _XAVR_PortD::Init(INP_PORTD_BITS, OUT_PORTD_BITS);
      _XAVR_PortE::Init(INP_PORTE_BITS, OUT_PORTE_BITS);
      _XAVR_PortF::Init(INP_PORTF_BITS, OUT_PORTF_BITS);
      _XAVR_PortG::Init(INP_PORTG_BITS, OUT_PORTG_BITS);
      _XAVR_PortH::Init(INP_PORTH_BITS, OUT_PORTH_BITS);
    }

    static inline void CheckAllAreOutputPins () _XAVR_INLINE
    {
      A0::CheckIsOutputPin(); A1::CheckIsOutputPin(); A2::CheckIsOutputPin();
      A3::CheckIsOutputPin(); A4::CheckIsOutputPin(); A5::CheckIsOutputPin();
      A6::CheckIsOutputPin(); A7::CheckIsOutputPin(); A8::CheckIsOutputPin();
    }

    static inline void CheckAllAreInputPins () _XAVR_INLINE
    {
      A0::CheckIsInputPin(); A1::CheckIsInputPin(); A2::CheckIsInputPin();
      A3::CheckIsInputPin(); A4::CheckIsInputPin(); A5::CheckIsInputPin();
      A6::CheckIsInputPin(); A7::CheckIsInputPin(); A8::CheckIsInputPin();
    }

    static inline void Set() _XAVR_INLINE
    {
      CheckAllAreOutputPins();
      _XAVR_PortA::AddM(OUT_PORTA_BITS);
      _XAVR_PortB::AddM(OUT_PORTB_BITS);
      _XAVR_PortC::AddM(OUT_PORTC_BITS);
      _XAVR_PortD::AddM(OUT_PORTD_BITS);
      _XAVR_PortE::AddM(OUT_PORTE_BITS);
      _XAVR_PortF::AddM(OUT_PORTF_BITS);
      _XAVR_PortG::AddM(OUT_PORTG_BITS);
      _XAVR_PortH::AddM(OUT_PORTH_BITS);
    }

    static inline void Clear() _XAVR_INLINE
    {
      CheckAllAreOutputPins();
      _XAVR_PortA::DelM(OUT_PORTA_BITS);
      _XAVR_PortB::DelM(OUT_PORTB_BITS);
      _XAVR_PortC::DelM(OUT_PORTC_BITS);
      _XAVR_PortD::DelM(OUT_PORTD_BITS);
      _XAVR_PortE::DelM(OUT_PORTE_BITS);
      _XAVR_PortF::DelM(OUT_PORTF_BITS);
      _XAVR_PortG::DelM(OUT_PORTG_BITS);
      _XAVR_PortH::DelM(OUT_PORTH_BITS);
    }

    static inline void Select (bool s0, bool s1 = false, bool s2 = false,
        bool s3 = false, bool s4 = false, bool s5 = false, bool s6 = false,
        bool s7 = false, bool s8 = false) _XAVR_INLINE
    {
      CheckAllAreOutputPins();

      #define _XAVR_SELECT_PINS_SET(N) \
          (s0 ? A0::OUT_PORT ## N ## _BIT : 0) | (s1 ? A1::OUT_PORT ## N ## _BIT : 0) | \
          (s2 ? A2::OUT_PORT ## N ## _BIT : 0) | (s3 ? A3::OUT_PORT ## N ## _BIT : 0) | \
          (s4 ? A4::OUT_PORT ## N ## _BIT : 0) | (s5 ? A5::OUT_PORT ## N ## _BIT : 0) | \
          (s6 ? A6::OUT_PORT ## N ## _BIT : 0) | (s7 ? A7::OUT_PORT ## N ## _BIT : 0) | \
          (s8 ? A8::OUT_PORT ## N ## _BIT : 0)

      #define _XAVR_SELECT_PINS_CLR(N) \
          (!s0 ? A0::OUT_PORT ## N ## _BIT : 0) | (!s1 ? A1::OUT_PORT ## N ## _BIT : 0) | \
          (!s2 ? A2::OUT_PORT ## N ## _BIT : 0) | (!s3 ? A3::OUT_PORT ## N ## _BIT : 0) | \
          (!s4 ? A4::OUT_PORT ## N ## _BIT : 0) | (!s5 ? A5::OUT_PORT ## N ## _BIT : 0) | \
          (!s6 ? A6::OUT_PORT ## N ## _BIT : 0) | (!s7 ? A7::OUT_PORT ## N ## _BIT : 0) | \
          (!s8 ? A8::OUT_PORT ## N ## _BIT : 0)

      uint8_t sa = _XAVR_SELECT_PINS_SET(A);
      uint8_t ca = _XAVR_SELECT_PINS_CLR(A);
      uint8_t sb = _XAVR_SELECT_PINS_SET(B);
      uint8_t cb = _XAVR_SELECT_PINS_CLR(B);
      uint8_t sc = _XAVR_SELECT_PINS_SET(C);
      uint8_t cc = _XAVR_SELECT_PINS_CLR(C);
      uint8_t sd = _XAVR_SELECT_PINS_SET(D);
      uint8_t cd = _XAVR_SELECT_PINS_CLR(D);
      uint8_t se = _XAVR_SELECT_PINS_SET(E);
      uint8_t ce = _XAVR_SELECT_PINS_CLR(E);
      uint8_t sf = _XAVR_SELECT_PINS_SET(F);
      uint8_t cf = _XAVR_SELECT_PINS_CLR(F);
      uint8_t sg = _XAVR_SELECT_PINS_SET(G);
      uint8_t cg = _XAVR_SELECT_PINS_CLR(G);
      uint8_t sh = _XAVR_SELECT_PINS_SET(H);
      uint8_t ch = _XAVR_SELECT_PINS_CLR(H);

      _XAVR_PortA::Select (sa, ca);
      _XAVR_PortB::Select (sb, cb);
      _XAVR_PortC::Select (sc, cc);
      _XAVR_PortD::Select (sd, cd);
      _XAVR_PortE::Select (se, ce);
      _XAVR_PortF::Select (sf, cf);
      _XAVR_PortG::Select (sg, cg);
      _XAVR_PortH::Select (sh, ch);
    }

    static inline void Toggle() _XAVR_INLINE
    {
      CheckAllAreOutputPins();
      _XAVR_PortA::XorM(OUT_PORTA_BITS);
      _XAVR_PortB::XorM(OUT_PORTB_BITS);
      _XAVR_PortC::XorM(OUT_PORTC_BITS);
      _XAVR_PortD::XorM(OUT_PORTD_BITS);
      _XAVR_PortE::XorM(OUT_PORTE_BITS);
      _XAVR_PortF::XorM(OUT_PORTF_BITS);
      _XAVR_PortG::XorM(OUT_PORTG_BITS);
      _XAVR_PortH::XorM(OUT_PORTH_BITS);
    }

    static inline bool AnyOn() _XAVR_INLINE
    {
      CheckAllAreInputPins();
      return INP_PORTA_BITS != 0 && _XAVR_PortA::Get() & INP_PORTA_BITS ||
          INP_PORTB_BITS != 0 && _XAVR_PortB::Get() & INP_PORTB_BITS ||
          INP_PORTC_BITS != 0 && _XAVR_PortC::Get() & INP_PORTC_BITS ||
          INP_PORTD_BITS != 0 && _XAVR_PortD::Get() & INP_PORTD_BITS ||
          INP_PORTE_BITS != 0 && _XAVR_PortE::Get() & INP_PORTE_BITS ||
          INP_PORTF_BITS != 0 && _XAVR_PortF::Get() & INP_PORTF_BITS ||
          INP_PORTG_BITS != 0 && _XAVR_PortG::Get() & INP_PORTG_BITS ||
          INP_PORTH_BITS != 0 && _XAVR_PortH::Get() & INP_PORTH_BITS;
    }

    static inline bool AllOn() _XAVR_INLINE
    {
      CheckAllAreInputPins();
      return (INP_PORTA_BITS == 0 || (_XAVR_PortA::Get() & INP_PORTA_BITS) == INP_PORTA_BITS) &&
          (INP_PORTB_BITS == 0 || (_XAVR_PortB::Get() & INP_PORTB_BITS) == INP_PORTB_BITS) &&
          (INP_PORTC_BITS == 0 || (_XAVR_PortC::Get() & INP_PORTC_BITS) == INP_PORTC_BITS) &&
          (INP_PORTD_BITS == 0 || (_XAVR_PortD::Get() & INP_PORTD_BITS) == INP_PORTD_BITS) &&
          (INP_PORTE_BITS == 0 || (_XAVR_PortE::Get() & INP_PORTE_BITS) == INP_PORTE_BITS) &&
          (INP_PORTF_BITS == 0 || (_XAVR_PortF::Get() & INP_PORTF_BITS) == INP_PORTF_BITS) &&
          (INP_PORTG_BITS == 0 || (_XAVR_PortG::Get() & INP_PORTG_BITS) == INP_PORTG_BITS) &&
          (INP_PORTH_BITS == 0 || (_XAVR_PortH::Get() & INP_PORTH_BITS) == INP_PORTH_BITS);
    }
  };

  template <class LA, class LB, class LC, class LD, class LE, class LF, class LG>
  class Display7
  {
  public:
    typedef Pins<LA, LB, LC, LD, LE, LF, LG> LEDs;

    static inline void Init () _XAVR_INLINE
    {
      LEDs::Init();
      Clear();
    }

    static inline void Clear () _XAVR_INLINE
    {
      Set (0xff);
    }

    static void Set (uint8_t n)
    {
      if ( n == 0 )
        {
          LEDs::Select (1, 1, 1, 1, 1, 1, 0);
        }
      else if ( n == 1 )
        {
          LEDs::Select (0, 1, 1, 0, 0, 0, 0);
        }
      else if ( n  == 2 )
        {
          LEDs::Select (1, 1, 0, 1, 1, 0, 1);
        }
      else if ( n  == 3 )
        {
          LEDs::Select (1, 1, 1, 1, 0, 0, 1);
        }
      else if ( n  == 4 )
        {
          LEDs::Select (0, 1, 1, 0, 0, 1, 1);
        }
      else if ( n  == 5 )
        {
          LEDs::Select (1, 0, 1, 1, 0, 1, 1);
        }
      else if ( n  == 6 )
        {
          LEDs::Select (1, 0, 1, 1, 1, 1, 1);
        }
      else if ( n  == 7 )
        {
          LEDs::Select (1, 1, 1, 0, 0, 0, 0);
        }
      else if ( n  == 8 )
        {
          LEDs::Set();
        }
      else if ( n  == 9 )
        {
          LEDs::Select (1, 1, 1, 1, 0, 1, 1);
        }
      else
        {
          LEDs::Clear();
        }
    }

  private:
    Display7();
  };

#if 0
  template<class P>
  struct PcInterrupt
  {
    typedef P PIN;

    static inline void Init () _XAVR_INLINE
    {
    }

    static inline void CheckIsInterrupt () _XAVR_INLINE
    {
      _xavr_user_error<PIN::PCINT == 0xff>::PC_interrupt_not_allowed_for_this_pin();
    }
  };
#endif
};

#if 0
#define XAVR_INTERRUPT(NAME, TYPE) \
  typedef TYPE NAME; \
  extern "C" void __vector_XAVR_INTERRUPT_ ## NAME ## _HDR  () __attribute__ ((naked, signal)); \
  extern "C" void __vector_XAVR_INTERRUPT_ ## NAME ## _HDR  () \
  { \
    TYPE::CheckIsInterrupt(); \
    TYPE::PIN::CheckIsInputPin(); \
    TYPE::PIN::_XAVR_InjectPCInterruptCode(); \
  } \
  extern "C" void __vector_XAVR_INTERRUPT_ ## NAME ## _IMPL  () __attribute__ ((signal)); \
  extern "C" void __vector_XAVR_INTERRUPT_ ## NAME ## _IMPL  ()
#endif

#endif // !__xavr_h__

